//(c) 2016 Dmitry Grinberg   http://dmitry.gr   me@dmitry.gr   see LICENSE.txt

#include <stdbool.h>
#include <stdint.h>
#include <avr/wdt.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include <util/delay.h>
#include <avr/pgmspace.h>
#include <stdio.h>
#include "flash.h"
#include "usb.h"
#include "hid.h"
#include "hw.h"



#define VENDOR_ID	((((uint16_t)'G') << 8) + (uint8_t)'D') 
#define PRODUCT_ID	((((uint16_t)'U') << 8) + (uint8_t)'g') 
#define POWER_NEED	500	//in milliamps



#define STRING_ID_LANGUAGE	0
#define STRING_ID_MANUF		1
#define STRING_ID_PRODUCT	2
#define STRING_ID_SERIAL	3

static const PROGMEM uint16_t usbPrvStrLang[] = {0x0409};
static const PROGMEM uint16_t usbPrvStrManuf[] = {'d','m','i','t','r','y','.','G','R'};
static const PROGMEM uint16_t usbPrvStrProd[] = {'K','L','N','8','9','B',' ','u','p','d','a','t','e','r'};
static uint16_t usbPrvStrSerial[20];	//note: this is in RAM, so we can edit it :)


#define CONFIG_NUM_MAIN		1
#define CONFIG_NUM_LAST		2

#define IFACE_NUM_HID_DATA	0
#define IFACE_NUM_LAST		1

#define ENDPT_NUM_HID_DATA_TX	1


static void init(){
	
	//wdt
	{		
		cli();
		wdt_reset();
		wdt_disable();
	}
	
	//USB PLL & power
	{
		UHWCON = 1;
		
		PLLCSR = 0b00011000;	//PLL set for external 16MHz crystal
		PLLCSR = 0b00011010;	//PLL set for external 16MHz crystal and on
		while(!(PLLCSR & 1));
	}
	
	extHwInit();
	flInit();
}



static uint8_t idleRates[IFACE_NUM_LAST];
static volatile bool mHaveRx = false, mNeedSendAck = false;
static volatile uint16_t mNeedRxLeft = 0, mHaveRxNow = 0;


/////////////////// main



USB_DEVICE_DESCR(usbPrvDescrDevice, VENDOR_ID, PRODUCT_ID, STRING_ID_MANUF, STRING_ID_PRODUCT, STRING_ID_SERIAL, CONFIG_NUM_LAST - 1)
	
USB_CONFIG_DESCR_START(usbPrvDescrDeviceConfig, 0, CONFIG_NUM_MAIN, USB_HID_DESCR_LEN(1), IFACE_NUM_LAST, 1 << 7, POWER_NEED)

	//let's add a data interface
	
	#define HID_ENDPT	ENDPT_NUM_HID_DATA_TX
	#define IFACE_NUMBER	IFACE_NUM_HID_DATA
	#define HID_POLL_INT	1
		HID_DESCRIPTOR_DATA()
	#undef HID_ENDPT
	#undef IFACE_NUMBER
	#undef HID_POLL_INT

USB_CONFIG_DESCR_END


REQ_DESCR_HANDLER_START
	
	REQ_DESCR_DEVICE(usbPrvDescrDevice)
	
	REQ_DESCR_CONFIG(0, usbPrvDescrDeviceConfig)
	
	REQ_DESCR_HID_DESCR_START
	
		HID_DESCR_HID_DESCR_EACH(IFACE_NUM_HID_DATA, usbPrvDescrDeviceConfig + USB_CFG_DESCR_LEN + 3*USB_HID_DESCR_LEN(1))
		
	REQ_DESCR_HID_DESCR_END
	
	REQ_DESCR_HID_REP_START
	
		REQ_DESCR_HID_REP_EACH(IFACE_NUM_HID_DATA, hidGetDescrData)
		
	REQ_DESCR_HID_REP_END
	
	REQ_DESCR_STRINGS_START
	
		REQ_DESCR_ADD_STRING_ROM(STRING_ID_LANGUAGE, usbPrvStrLang)
		REQ_DESCR_ADD_STRING_ROM(STRING_ID_MANUF, usbPrvStrManuf)
		REQ_DESCR_ADD_STRING_ROM(STRING_ID_PRODUCT, usbPrvStrProd)
		REQ_DESCR_ADD_STRING_RAM(STRING_ID_SERIAL, usbPrvStrSerial)
	
	REQ_DESCR_STRINGS_END

REQ_DESCR_HANDLER_END


USB_CUSTOM_EP0_START

	REQ_CUSTOM_EP0_HID_GET_REP_START
	
		REQ_CUSTOM_EP0_HID_GET_REP_EACH(IFACE_NUM_HID_DATA, NULL, 0)
	
	REQ_CUSTOM_EP0_HID_GET_REP_END
	
	REQ_CUSTOM_EP0_HID_GET_IDLE_START
		
		REQ_CUSTOM_EP0_HID_GET_IDLE_EACH(IFACE_NUM_HID_DATA, idleRates)
	
	REQ_CUSTOM_EP0_HID_GET_IDLE_END
	
	REQ_CUSTOM_EP0_HID_SET_IDLE_START
		
		REQ_CUSTOM_EP0_HID_SET_IDLE_EACH(IFACE_NUM_HID_DATA, idleRates)
	
	REQ_CUSTOM_EP0_HID_SET_IDLE_END
	
	//set feature report request?
	else if(BM_REQ_IS_CLS(bmRequestType) && bRequest == HID_SET_REPORT && wValueH == 2 /* hid report */ && wValueL == 0 /* feature idx 0 */) {
		
		//for our interface?
		if(wIndexH == 0 && wIndexL == IFACE_NUM_HID_DATA) {
			
			//we're not too busy?
			if (!mNeedRxLeft && !mHaveRxNow && !mNeedSendAck) {
			
				mNeedRxLeft = (((uint16_t)wLengthH) << 8) | wLengthL;
				mHaveRxNow = 0;
				sendZlpAckOnSuccess = false;
				return true;
			}
		}
		
		return false;
	}
	
USB_CUSTOM_EP0_END

bool usbCbkSetConfig(uint8_t cfg){

	if((cfg &~ CFG_NUM_OR_AFTER) != 1) return false;	//we only have one config, index one
	
	if(cfg & CFG_NUM_OR_AFTER){
		
		HID_ENDPT_CONFIG(ENDPT_NUM_HID_DATA_TX);
	}
	return true;
}

static char memcpycmp(void* __restrict dstP, const void* __restrict srcP, uint8_t len)
{
	uint8_t *dst = dstP;
	const uint8_t *src = srcP;
	char diff = 0;
	
	while(len--){
		if(*src != *dst) diff = 1;
		*dst++ = *src++;
	}
	
	return diff;
}

static void getInfo(struct CommsPacket *dst)
{
	dst->info.version = 0x00010000;
	dst->info.compatible = flCheckCompatible();
	if (dst->info.compatible) {
		flRead(0x42, (uint8_t*)&dst->info.key, sizeof(dst->info.key));
		flRead(0x46, (uint8_t*)&dst->info.magicval, sizeof(dst->info.magicval));
	}
	flGetGeometryInfo(&dst->info.flashSz, &dst->info.flashErzBlkSz);
}

static void __attribute__((noreturn)) main_master(void)
{
	struct CommsPacket bufs[2];
	struct CommsPacket *bufRx = bufs + 0;
	struct CommsPacket *bufTx = bufs + 1;
	bool haveDataToTx = false;
	uint16_t bytesAvail;
	
	while(1){
		
		if (mNeedRxLeft && usbEpHasData(0)) {
			
			uint8_t now = (mNeedRxLeft > USB_EP0_SIZE) ? USB_EP0_SIZE : mNeedRxLeft;
			
			usbEpRxBytes(0, ((uint8_t*)bufRx) + mHaveRxNow, now);
			mHaveRxNow += now;
			mNeedRxLeft -= now;
			
			if (!mNeedRxLeft)
				mNeedSendAck = true;
		}
		
		if (mNeedSendAck && usbEp0canSend()) {
			
			mNeedSendAck = false;
			usbEpDoSend(0, 0, 0);
		}
		
		if(usbEpCanSend(ENDPT_NUM_HID_DATA_TX) && haveDataToTx){
			usbEpDoSend(ENDPT_NUM_HID_DATA_TX, bufTx, sizeof(*bufTx));
			haveDataToTx = false;
		}
		
		if (mHaveRxNow && !mNeedRxLeft) {
		
			struct CommsPacket *t = bufTx;
			bufTx = bufRx;
			bufRx = t;
			
			mHaveRxNow = 0;
		
			if (bufTx->crc != commsPacketCalc(bufTx))
				continue;
			
			switch (bufTx->cmd) {
				case PACKET_TYPE_GET_INFO:
					getInfo(bufTx);
					break;
					
				case PACKET_TYPE_DO_ERASE:
					bufTx->success = flErase(bufTx->erase.addr);
					break;
				
				case PACKET_TYPE_DO_WRITE:
					bufTx->success = flWrite(bufTx->data.addr, bufTx->data.bytes, COMMS_DATA_TX_SZ);
					break;
				
				case PACKET_TYPE_DO_READ:
					if (!flRead(bufTx->data.addr, bufTx->data.bytes, COMMS_DATA_TX_SZ))
						bufTx->data.addr = 0xfffffffful;	//we indicate error thusly
					break;
				
				case PACKET_TYPE_VPP_CTRL:
					extHwVpp(bufTx->on);
					if (bufTx->on)
						_delay_ms(100);		//stabilization
					break;
				
				default:
					continue;
			}
			
			bufTx->crc = commsPacketCalc(bufTx);
			haveDataToTx = true;
		}
	}
}

static void snumGen(void)			//do not run with interrupts on
{
	const uint8_t snum_addr_start = 0x0E;
	uint8_t i, v;
	
	for(i = 0; i < sizeof(usbPrvStrSerial) /sizeof(*usbPrvStrSerial) / 2; i++) {
		
		asm volatile(
			"mov r30, %1	\n\t"
			"mov r31, r1	\n\t"
			"out 0x37, %2	\n\t"
			"lpm %0, Z		\n\t"
			:"=r"(v)
			:"r"(i + snum_addr_start), "r"(0b00100001)
			:"r30", "r31", "cc"
		);
		asm volatile("":::"memory");
		usbPrvStrSerial[i * 2 + 0] = ((v >> 4) >= 10) ? ((v >> 4) + 'a' - 10) : ((v >> 4) + '0');
		usbPrvStrSerial[i * 2 + 1] = ((v & 15) >= 10) ? ((v & 15) + 'a' - 10) : ((v & 15) + '0');
	}
	SPMCSR = 0;
}

int __attribute__((noreturn)) main()
{	
	init();
	snumGen();
	
	usbInit(false, false);
	usbAttachControl(false);
	_delay_ms(1000);

	
	sei();
	extHwLed(false);
	usbAttachControl(true);
	
	
	while(!usbIsConfigured());
	
	
	main_master();
}

